from typing import *
import asyncio
import os
import json
import inspect
from datetime import datetime, timezone as tz
from anchorpy import Idl, Program, Provider, Context, Wallet
from anchorpy.error import AccountDoesNotExistError
from solana.publickey import PublicKey
from solana.keypair import Keypair
from solana.transaction import TransactionInstruction, Transaction, TransactionSignature
from solana.rpc.commitment import Commitment, Finalized
from solana.rpc.async_api import AsyncClient
from solana.rpc.types import TxOpts
from solana.sysvar import SYSVAR_RENT_PUBKEY
from solana.system_program import SYS_PROGRAM_ID
from spl.token.instructions import get_associated_token_address
from spl.token.constants import TOKEN_PROGRAM_ID

from . import util, types, config
from .config import configs, Config
from .types import (
    Side,
    OrderType,
    CollateralInfo,
    FundingInfo,
    MarketInfo,
    PositionInfo,
)
from .dex import Market, Orderbook, Order

T = TypeVar("T")


def GenIxDispatch(cls):
    """Decorator for use with the `Zo` class.

    This decorator finds all methods ending in `_ix` on the class,
    and adds a trivial wrapper to dispatch the instruction using `Zo.send`.
    The `__doc__` is moved to the generated method.
    """
    for n, f in inspect.getmembers(cls, predicate=inspect.isfunction):
        if n.startswith("_") or not n.endswith("_ix"):
            continue

        async def g(self, *a, **kw):
            # If you're looking for the source, see the source
            # for the `_ix` variant of this method.
            return await self.send(f(*a, **kw))

        name = n[:-3]
        g.__name__ = name
        g.__qualname__ = f.__qualname__[:-3]
        g.__doc__ = inspect.getdoc(f)
        g.__signature__ = inspect.signature(f).replace(
            return_annotation=TransactionSignature
        )
        g.__annotations__ = inspect.get_annotations(f)  # Copies.
        g.__annotations__["return"] = TransactionSignature

        # The docs are intended for the non '_ix' variant, so modify them.
        f.__doc__ = f"See `Zo.{name}`."

        setattr(cls, name, g)

    return cls


class ZoIndexer(Generic[T]):
    def __init__(self, d: dict[str, T], m: Callable[[str | int | PublicKey], str]):
        self.d = d
        self.m = m

    def __repr__(self):
        return self.d.__repr__()

    def __iter__(self):
        return self.d.items().__iter__()

    def __len__(self):
        return len(self.d)

    def __getitem__(self, i: str | int | PublicKey) -> T:
        return self.d[self.m(i)]


@GenIxDispatch
class Zo:
    __program: Program
    __config: Config

    __markets: dict[str, MarketInfo]
    __collaterals: dict[str, CollateralInfo]
    __orderbook: dict[str, Orderbook]
    __balance: dict[str, float]
    __position: dict[str, PositionInfo]

    __dex_markets: dict[str, Market]
    __orders: dict[str, list[Order]]

    __markets_map: dict[str | int | PublicKey, str]
    __collaterals_map: dict[str | int | PublicKey, str]

    _zo_state: Any
    _zo_state_signer: PublicKey
    _zo_cache: Any
    _zo_margin: Any
    _zo_margin_key: None | PublicKey
    _zo_control: Any

    def __init__(
        self,
        *,
        _program,
        _config,
        _state,
        _state_signer,
        _margin,
        _margin_key,
    ):
        self.__program = _program
        self.__config = _config
        self._zo_state = _state
        self._zo_state_signer = _state_signer
        self._zo_margin = _margin
        self._zo_margin_key = _margin_key

    @classmethod
    async def new(
        cls,
        *,
        cluster: Literal["devnet", "mainnet"],
        payer: Keypair | None = None,
        url: str | None = None,
        load_margin: bool = True,
        create_margin: bool = True,
        tx_opts: TxOpts = TxOpts(
            max_retries=None,
            preflight_commitment=Finalized,
            skip_confirmation=False,
            skip_preflight=False,
        ),
    ):
        """Create a new client instance.

        Args:
            cluster: Which cluster to connect to.
            payer: The transaction payer and margin owner. Defaults to
                the local transaction payer.
            url: URL for the RPC endpoint.
            load_margin: Whether to load the associated margin account.
                If `False`, any transaction requiring a margin will fail.
            create_margin: Whether to create the associated margin
                account if it doesn't already exist.
            tx_opts: The transaction options.
        """

        if cluster not in configs.keys():
            raise TypeError(f"`cluster` must be one of: {configs.keys()}")

        config = configs[cluster]

        if url is None:
            url = config.CLUSTER_URL

        idl_path = os.path.join(os.path.dirname(__file__), "idl.json")
        with open(idl_path) as f:
            raw_idl = json.load(f)

        idl = Idl.from_json(raw_idl)
        conn = AsyncClient(url)
        wallet = Wallet(payer) if payer is not None else Wallet.local()
        provider = Provider(conn, wallet, opts=tx_opts)
        program = Program(idl, config.ZO_PROGRAM_ID, provider=provider)

        state = await program.account["State"].fetch(config.ZO_STATE_ID)
        state_signer, state_signer_nonce = util.state_signer_pda(
            state=config.ZO_STATE_ID, program_id=config.ZO_PROGRAM_ID
        )

        if state.signer_nonce != state_signer_nonce:
            raise ValueError(
                f"Invalid state key ({config.ZO_STATE_ID}) for program id ({config.ZO_PROGRAM_ID})"
            )

        margin = None
        margin_key = None

        if load_margin:
            margin_key, nonce = util.margin_pda(
                owner=wallet.public_key,
                state=config.ZO_STATE_ID,
                program_id=config.ZO_PROGRAM_ID,
            )
            try:
                margin = await program.account["Margin"].fetch(margin_key)
            except AccountDoesNotExistError as e:
                if not create_margin:
                    raise e

                await util.create_margin(
                    program=program,
                    state=config.ZO_STATE_ID,
                    key=margin_key,
                    nonce=nonce,
                )
                margin = await program.account["Margin"].fetch(margin_key)

        zo = cls(
            _config=config,
            _program=program,
            _state=state,
            _state_signer=state_signer,
            _margin=margin,
            _margin_key=margin_key,
        )
        await zo.refresh(commitment=Finalized)
        return zo

    @property
    def program(self) -> Program:
        return self.__program

    @property
    def provider(self) -> Provider:
        return self.program.provider

    @property
    def connection(self) -> AsyncClient:
        return self.provider.connection

    @property
    def wallet(self) -> Wallet:
        return self.provider.wallet

    @property
    def collaterals(self):
        """List of collaterals and their metadata."""
        return ZoIndexer(self.__collaterals, lambda k: self.__collaterals_map[k])

    @property
    def markets(self):
        """List of collaterals and markets metadata."""
        return ZoIndexer(self.__markets, lambda k: self.__markets_map[k])

    @property
    def orderbook(self):
        """Current state of the orderbook."""
        return ZoIndexer(self.__orderbook, lambda k: self.__markets_map[k])

    @property
    def balance(self):
        """Current account balance."""
        return ZoIndexer(self.__balance, lambda k: self.__collaterals_map[k])

    @property
    def position(self):
        """Current position."""
        return ZoIndexer(self.__position, lambda k: self.__markets_map[k])

    @property
    def orders(self):
        """Currently active orders."""
        return ZoIndexer(self.__orders, lambda k: self.__markets_map[k])

    def _get_open_orders_info(self, key: int | str, /):
        if isinstance(key, str):
            for k, v in self.__markets_map.items():
                if v == key and isinstance(k, int):
                    key = k
                    break
            else:
                ValueError("")
        o = self._zo_control.open_orders_agg[key]
        return o if o.key != PublicKey(0) else None

    def __reload_collaterals(self):
        map = {}
        collaterals = {}

        for i, c in enumerate(self._zo_state.collaterals):
            if c.mint == PublicKey(0):
                break

            symbol = util.decode_symbol(c.oracle_symbol)
            map[symbol] = symbol
            map[i] = symbol
            map[c.mint] = symbol

            collaterals[symbol] = CollateralInfo(
                mint=c.mint,
                oracle_symbol=symbol,
                decimals=c.decimals,
                weight=c.weight,
                liq_fee=c.liq_fee,
                is_borrowable=c.is_borrowable,
                optimal_util=c.optimal_util,
                optimal_rate=c.optimal_rate,
                max_rate=c.max_rate,
                og_fee=c.og_fee,
                is_swappable=c.is_swappable,
                serum_open_orders=c.serum_open_orders,
                max_deposit=c.max_deposit,
                dust_threshold=c.dust_threshold,
                vault=self._zo_state.vaults[i],
            )

        self.__collaterals_map = map
        self.__collaterals = collaterals

    def __reload_markets(self):
        map = {}
        markets = {}

        for i, m in enumerate(self._zo_state.perp_markets):
            if m.dex_market == PublicKey(0):
                break

            symbol = util.decode_symbol(m.symbol)

            if symbol == "LUNA-PERP":
                continue

            map[symbol] = symbol
            map[i] = symbol
            map[m.dex_market] = symbol

            oracle = None
            for o in reversed(self._zo_cache.oracles):
                if util.decode_symbol(m.oracle_symbol) == util.decode_symbol(o.symbol):
                    oracle = o
                    break
            else:
                raise IndexError(f"oracle for market {symbol} not found")

            mark = self._zo_cache.marks[i]

            price_adj = 10 ** (m.asset_decimals - 6)
            index_price = util.decode_wrapped_i80f48(oracle.price) * price_adj
            mark_price = util.decode_wrapped_i80f48(mark.price) * price_adj

            if types.perp_type_to_str(m.perp_type, program=self.program) == "square":
                index_price = index_price**2 / m.strike

            funding_sample_start = datetime.fromtimestamp(
                mark.twap.last_sample_start_time, tz=tz.utc
            )
            cumul_avg = util.decode_wrapped_i80f48(mark.twap.cumul_avg)
            if abs(cumul_avg) == 0 or funding_sample_start.minute == 0:
                funding_info = None
            else:
                daily_funding = cumul_avg / funding_sample_start.minute
                funding_info = FundingInfo(
                    daily=daily_funding,
                    hourly=daily_funding / 24,
                    apr=daily_funding * 100 * 365,
                )

            markets[symbol] = MarketInfo(
                address=m.dex_market,
                symbol=symbol,
                oracle_symbol=util.decode_symbol(m.oracle_symbol),
                perp_type=types.perp_type_to_str(m.perp_type, program=self.program),
                base_decimals=m.asset_decimals,
                base_lot_size=m.asset_lot_size,
                quote_decimals=6,
                quote_lot_size=m.quote_lot_size,
                strike=m.strike,
                base_imf=m.base_imf,
                liq_fee=m.liq_fee,
                index_price=index_price,
                mark_price=mark_price,
                funding_sample_start_time=funding_sample_start,
                funding_info=funding_info,
            )

        self.__markets_map = map
        self.__markets = markets

    def __reload_balances(self):
        if self._zo_margin is None:
            return

        balances = {}
        for i, c in enumerate(self._zo_margin.collateral):
            if i not in self.__collaterals_map:
                break

            decimals = self.collaterals[i].decimals
            c = util.decode_wrapped_i80f48(c)
            m = self._zo_cache.borrow_cache[i]
            m = m.supply_multiplier if c >= 0 else m.borrow_multiplier
            m = util.decode_wrapped_i80f48(m)

            balances[self.__collaterals_map[i]] = util.small_to_big_amount(
                c * m, decimals=decimals
            )

        self.__balance = balances

    def __reload_positions(self):
        if self._zo_margin is None:
            return

        positions = {}
        for s, m in self.markets:
            if (oo := self._get_open_orders_info(s)) is not None:
                positions[s] = PositionInfo(
                    size=util.small_to_big_amount(
                        abs(oo.pos_size), decimals=m.base_decimals
                    ),
                    entry_value=util.small_to_big_amount(
                        abs(oo.native_pc_total), decimals=m.quote_decimals
                    ),
                    realized_pnl=util.small_to_big_amount(
                        oo.realized_pnl, decimals=m.base_decimals
                    ),
                    funding_index=util.small_to_big_amount(
                        oo.funding_index, decimals=m.quote_decimals
                    ),
                    side="long" if oo.pos_size >= 0 else "short",
                )
            else:
                positions[s] = PositionInfo(
                    size=0, entry_value=0, realized_pnl=0, funding_index=1, side="long"
                )

        self.__position = positions
        pass

    async def __reload_dex_markets(self, *, commitment: None | Commitment = None):
        ks = [
            m.dex_market
            for m in self._zo_state.perp_markets
            if m.dex_market in self.__markets_map
        ]
        res: Any = await self.connection.get_multiple_accounts(
            ks, encoding="base64", commitment=commitment
        )
        res = res["result"]["value"]
        self.__dex_markets = {
            self.__markets_map[ks[i]]: Market.from_base64(res[i]["data"][0])
            for i in range(len(ks))
        }

    async def refresh_orders(self, *, commitment: None | Commitment = None):
        """Refresh only the orderbook and orders."""
        ss = []
        ks = []
        for s, mkt in self.__dex_markets.items():
            ss.append(s)
            ks.extend((mkt.bids, mkt.asks))

        res: Any = await self.connection.get_multiple_accounts(
            ks, encoding="base64", commitment=commitment
        )
        res = res["result"]["value"]
        orders = self._zo_margin and {}
        orderbook = {}

        for i, s in enumerate(ss):
            mkt = self.__dex_markets[s]
            ob = mkt._decode_orderbook_from_base64(
                res[2 * i]["data"][0], res[2 * i + 1]["data"][0]
            )
            orderbook[s] = ob

            if self._zo_margin is not None:
                os = []
                for slab in [ob.bids, ob.asks]:
                    for o in slab:
                        if o.control == self._zo_margin.control:
                            os.append(o)
                orders[s] = os

        self.__orderbook = orderbook
        self.__orders = orders

    async def __refresh_margin(self, *, commitment: None | Commitment = None):
        if self._zo_margin_key is not None:
            self._zo_margin, self._zo_control = await asyncio.gather(
                self.program.account["Margin"].fetch(self._zo_margin_key, commitment),
                self.program.account["Control"].fetch(
                    self._zo_margin.control, commitment
                ),
            )

    async def refresh(self, *, commitment: Commitment = Finalized):
        """Refresh the loaded accounts to see updates."""
        self._zo_state, self._zo_cache, _ = await asyncio.gather(
            self.program.account["State"].fetch(self.__config.ZO_STATE_ID, commitment),
            self.program.account["Cache"].fetch(self._zo_state.cache, commitment),
            self.__refresh_margin(),
        )

        self.__reload_collaterals()
        self.__reload_markets()
        self.__reload_balances()
        self.__reload_positions()
        await self.__reload_dex_markets(commitment=commitment)
        await self.refresh_orders(commitment=commitment)

    async def send(
        self, *xs: TransactionInstruction | Transaction
    ) -> TransactionSignature:
        """Dispatch a set of instructions.

        Args:
            xs: The instruction or transaction to add.

        Returns:
            The transaction signature.
        """
        tx = Transaction()
        for x in xs:
            tx.add(x)
        return await self.provider.send(tx)

    def deposit_ix(
        self,
        amount: float,
        /,
        *,
        mint: PublicKey,
        repay_only: bool = False,
        token_account: None | PublicKey = None,
    ) -> TransactionInstruction:
        """Deposit collateral into the margin account.

        Args:
            amount: The amount to deposit, in big units (e.g.: 1.5 SOL, 0.5 BTC).
            mint: Mint of the collateral being deposited.
            repay_only: If true, will only deposit up to the amount borrowed.
            token_account: The token account to deposit from, defaulting to
                the associated token account.

        Returns:
            The transaction signature.
        """
        if token_account is None:
            token_account = get_associated_token_address(self.wallet.public_key, mint)

        decimals = self.collaterals[mint].decimals
        amount = util.big_to_small_amount(amount, decimals=decimals)

        return self.program.instruction["deposit"](
            repay_only,
            amount,
            ctx=Context(
                accounts={
                    "state": self.__config.ZO_STATE_ID,
                    "state_signer": self._zo_state_signer,
                    "cache": self._zo_state.cache,
                    "authority": self.wallet.public_key,
                    "margin": self._zo_margin_key,
                    "token_account": token_account,
                    "vault": self.collaterals[mint].vault,
                    "token_program": TOKEN_PROGRAM_ID,
                }
            ),
        )

    def withdraw_ix(
        self,
        amount: float,
        /,
        *,
        mint: PublicKey,
        allow_borrow: bool = False,
        token_account: None | PublicKey = None,
    ) -> TransactionInstruction:
        """Withdraw collateral from the margin account.

        Args:
            amount: The amount to withdraw, in big units (e.g.: 1.5 SOL, 0.5 BTC).
            mint: The mint of the collateral.
            allow_borrow: If true, will allow borrowing.
            token_account: If false, will only be able to withdraw up to
                the amount deposited. If false, amount parameter can be
                set to an arbitrarily large number to ensure that all
                deposits are fully withdrawn.

        Returns:
            The transaction signature.
        """

        if token_account is None:
            token_account = get_associated_token_address(self.wallet.public_key, mint)

        decimals = self.collaterals[mint].decimals
        amount = util.big_to_small_amount(amount, decimals=decimals)

        return self.program.instruction["withdraw"](
            allow_borrow,
            amount,
            ctx=Context(
                accounts={
                    "state": self.__config.ZO_STATE_ID,
                    "state_signer": self._zo_state_signer,
                    "cache": self._zo_state.cache,
                    "authority": self.wallet.public_key,
                    "margin": self._zo_margin_key,
                    "control": self._zo_margin.control,
                    "token_account": token_account,
                    "vault": self.collaterals[mint].vault,
                    "token_program": TOKEN_PROGRAM_ID,
                }
            ),
        )

    def place_order_ix(
        self,
        size: float,
        price: float,
        side: Side,
        *,
        symbol: str,
        order_type: OrderType,
        limit: int = 10,
        client_id: int = 0,
        max_ts: None | int = None,
    ) -> Transaction:
        """Place an order on the orderbook.

        Args:
            size: The maximum amount of big base units to buy or sell.
            price: The limit price in big quote units per big base
                units, e.g. 50000 USD/SOL.
            side: Whether to place a bid or an ask.
            symbol: The market symbol, e.g. "BTC-PERP".
            order_type: The order type.
            limit: If this order is taking, the limit sets the number of
                maker orders the fill will go through, until stopping and
                posting. If running into compute unit issues, then set
                this number lower.
            client_id: Used to tag an order with a unique id, which can
                be used to cancel this order through
                cancelPerpOrderByClientId. For optimal use, make sure
                all ids for every order is unique.
            max_ts: If the current on-chain Unix timestamp exceeds this
                value, then the order will not go through.

        Returns:
            The transaction signature.
        """

        mkt = self.__dex_markets[symbol]
        info = self.markets[symbol]
        is_long = side == "bid"
        price = util.price_to_lots(
            price,
            base_decimals=info.base_decimals,
            quote_decimals=info.quote_decimals,
            base_lot_size=info.base_lot_size,
            quote_lot_size=info.quote_lot_size,
        )
        order_type_: Any = types.order_type_from_str(order_type, program=self.program)
        taker_fee = config.taker_fee(info.perp_type)
        fee_multiplier = 1 + taker_fee if is_long else 1 - taker_fee
        base_qty = util.size_to_lots(
            size, decimals=info.base_decimals, lot_size=info.base_lot_size
        )
        quote_qty = round(price * fee_multiplier * base_qty * info.quote_lot_size)

        pre_ixs = []
        oo_key = None
        oo_info = self._get_open_orders_info(symbol)

        if oo_info is not None:
            oo_key = oo_info.key
        else:
            oo_key, _ = util.open_orders_pda(
                control=self._zo_margin.control,
                dex_market=info.address,
                program_id=self.program.program_id,
            )
            pre_ixs = [
                self.program.instruction["create_perp_open_orders"](
                    ctx=Context(
                        accounts={
                            "state": self.__config.ZO_STATE_ID,
                            "state_signer": self._zo_state_signer,
                            "authority": self.wallet.public_key,
                            "payer": self.wallet.public_key,
                            "margin": self._zo_margin_key,
                            "control": self._zo_margin.control,
                            "open_orders": oo_key,
                            "dex_market": info.address,
                            "dex_program": self.__config.ZO_DEX_ID,
                            "rent": SYSVAR_RENT_PUBKEY,
                            "system_program": SYS_PROGRAM_ID,
                        },
                    )
                )
            ]

        return self.program.instruction["place_perp_order_with_max_ts"](
            is_long,
            price,
            base_qty,
            quote_qty,
            order_type_,
            limit,
            client_id,
            max_ts if max_ts is not None else 2**63 - 1,
            ctx=Context(
                accounts={
                    "state": self.__config.ZO_STATE_ID,
                    "state_signer": self._zo_state_signer,
                    "cache": self._zo_state.cache,
                    "authority": self.wallet.public_key,
                    "margin": self._zo_margin_key,
                    "control": self._zo_margin.control,
                    "open_orders": oo_key,
                    "dex_market": info.address,
                    "req_q": mkt.req_q,
                    "event_q": mkt.event_q,
                    "market_bids": mkt.bids,
                    "market_asks": mkt.asks,
                    "dex_program": self.__config.ZO_DEX_ID,
                    "rent": SYSVAR_RENT_PUBKEY,
                },
                pre_instructions=pre_ixs,
            ),
        )

    def __cancel_order_ix(
        self,
        *,
        symbol: str,
        side: None | Side = None,
        order_id: None | int = None,
        client_id: None | int = None,
    ) -> TransactionInstruction:
        mkt = self.__dex_markets[symbol]
        oo = self._get_open_orders_info(symbol)

        if oo is None:
            raise ValueError("open orders account is uninitialized")

        return self.program.instruction["cancel_perp_order"](
            order_id,
            side == "bid",
            client_id,
            ctx=Context(
                accounts={
                    "state": self.__config.ZO_STATE_ID,
                    "cache": self._zo_state.cache,
                    "authority": self.wallet.public_key,
                    "margin": self._zo_margin_key,
                    "control": self._zo_margin.control,
                    "open_orders": oo.key,
                    "dex_market": mkt.own_address,
                    "market_bids": mkt.bids,
                    "market_asks": mkt.asks,
                    "event_q": mkt.event_q,
                    "dex_program": self.__config.ZO_DEX_ID,
                }
            ),
        )

    def cancel_order_by_order_id_ix(
        self, order_id: int, side: Side, *, symbol: str
    ) -> TransactionInstruction:
        """Cancel an order on the orderbook using the `order_id`.

        Args:
            order_id: The order id of the order to cancel. To get the
                order_id, see the `orders` field.
            side: Whether the order is a bid or an ask.
            symbol: The market symbol, e.g. "BTC-PERP".

        Returns:
            The transaction signature.
        """
        return self.__cancel_order_ix(symbol=symbol, order_id=order_id, side=side)

    def cancel_order_by_client_id_ix(
        self, client_id: int, *, symbol: str
    ) -> TransactionInstruction:
        """Cancel an order on the orderbook using the `client_id`.

        Args:
            client_id: The client id that was assigned to the order
                when it was placed..
            symbol: The market symbol, e.g. "BTC-PERP".

        Returns:
            The transaction signature.
        """
        return self.__cancel_order_ix(symbol=symbol, client_id=client_id)

    def close_position_ix(self, symbol: str, /, *, max_ts: int | None = None):
        """Close a position.

        Args:
            symbol: The markey symbol, e.g. "BTC-PERP".
            max_ts: See `Zo.place_order`.

        Returns:
            The transaction signature.
        """
        pos = self.position[symbol]
        side = "ask" if pos.side == "long" else "bid"
        price = 0 if side == "ask" else 2**63 - 1

        return self.place_order_ix(
            pos.size,
            price,
            side,
            symbol=symbol,
            order_type="reduceonlyioc",
            max_ts=max_ts,
        )



import base64
import enum
import struct
from typing import *
from solana.publickey import PublicKey

from . import util

i128 = NewType("i128", int)
u128 = NewType("u128", int)


def decode_field(ty, v):
    if ty == i128:
        return ty(int.from_bytes(v, "little", signed=True))
    elif ty == u128:
        return ty(int.from_bytes(v, "little", signed=False))
    else:
        return ty(v)


def strip_padding(b: bytes) -> bytes:
    if len(b) < 12 or b[:5] != b"serum" or b[-7:] != b"padding":
        raise ValueError("invalid buffer for dex struct")
    else:
        return b[5:-7]


def decode_namedtuple(cls, fmt: str, b: bytes):
    return cls._make(
        [
            decode_field(cls.__annotations__[f], g)
            for f, g in zip(cls._fields, struct.unpack(fmt, b))
        ]
    )


class AccountFlag(enum.IntFlag):
    INITIALIZED = 1 << 0
    MARKET = 1 << 1
    OPEN_ORDERS = 1 << 2
    REQUEST_QUEUE = 1 << 3
    EVENT_QUEUE = 1 << 4
    BIDS = 1 << 5
    ASKS = 1 << 6
    DISABLED = 1 << 7
    CLOSED = 1 << 8
    PERMISSIONED = 1 << 9


class Market(NamedTuple):
    account_flags: AccountFlag
    own_address: PublicKey
    quote_fees_accrued: int
    req_q: PublicKey
    event_q: PublicKey
    bids: PublicKey
    asks: PublicKey
    base_lot_size: int
    quote_lot_size: int
    fee_rate_bps: int
    referrer_rebates_accrued: int
    funding_index: i128
    last_updated: int
    strike: int
    perp_type: int
    base_decimals: int
    open_interest: int
    open_orders_authority: PublicKey
    prune_authority: PublicKey

    @property
    def quote_decimals(self):
        return 6

    @classmethod
    def from_bytes(cls, b: bytes):
        b = strip_padding(b)
        r = decode_namedtuple(cls, "<Q32sQ32s32s32s32s4Q16s5Q32s32s1032x", b)
        if (
            r.account_flags
            != AccountFlag.INITIALIZED | AccountFlag.MARKET | AccountFlag.PERMISSIONED
        ):
            raise ValueError("invalid account_flags for market")
        return r

    @classmethod
    def from_base64(cls, b: str):
        return cls.from_bytes(base64.b64decode(b))

    def _decode_orderbook_from_base64(self, bids: str, asks: str):
        return Orderbook(Slab.from_base64(bids), Slab.from_base64(asks), self)


class SlabNode:
    class Uninitialized(NamedTuple):
        pass

    class Inner(NamedTuple):
        prefix_len: int
        key: u128
        l: int
        r: int

    class Leaf(NamedTuple):
        owner_slot: int
        fee_tier: int
        key: u128
        control: PublicKey
        quantity: int
        client_order_id: int

    class Free(NamedTuple):
        next: int

    class LastFree(NamedTuple):
        pass

    @classmethod
    def _from_bytes(cls, b: bytes):
        tag, body = int.from_bytes(b[:4], "little", signed=False), b[4:]

        if tag < 0 or tag > 4:
            raise ValueError(f"invalid tag type '{tag}' for slab node")

        fmt, cons = [
            ("<68x", cls.Uninitialized),
            ("<I16sII40x", cls.Inner),
            ("<BBxx16s32sQQ", cls.Leaf),
            ("<I64x", cls.Free),
            ("<68x", cls.LastFree),
        ][tag]

        return decode_namedtuple(cons, fmt, body)


class Order(NamedTuple):
    owner_slot: int
    fee_tier: int
    order_id: u128
    control: PublicKey
    size_lots: int
    client_order_id: int
    price: float
    size: float
    side: Literal["bid", "ask"]


class Slab(NamedTuple):
    account_flags: AccountFlag
    bump_index: int
    free_list_len: int
    free_list_head: int
    root: int
    leaf_count: int
    nodes: list[SlabNode]

    @property
    def side(self) -> Literal["bid", "ask"]:
        return "bid" if AccountFlag.BIDS in self.account_flags else "ask"

    def __iter(self, *, ascending: bool) -> Generator[SlabNode.Leaf, None, None]:
        if self.leaf_count <= 0:
            return

        stack = [self.root]
        while len(stack) > 0:
            node = self.nodes[stack.pop()]
            if isinstance(node, SlabNode.Leaf):
                yield node
            elif isinstance(node, SlabNode.Inner):
                if ascending:
                    stack.extend((node.r, node.l))
                else:
                    stack.extend((node.l, node.r))

    def __iter__(self):
        return self.__iter(ascending=True)

    def __reversed__(self):
        return self.__iter(ascending=False)

    def min(self):
        return next(self.__iter(ascending=True))

    def max(self):
        return next(self.__iter(ascending=False))

    @classmethod
    def from_bytes(cls, b: bytes):
        b = strip_padding(b)
        head, tail = b[:40], b[40:]

        fs = list(struct.unpack("<QQQIIQ", head))
        fs[0] = decode_field(AccountFlag, fs[0])

        nodes = []
        for i in range(0, len(tail), 72):
            x = SlabNode._from_bytes(tail[i : i + 72])
            if isinstance(x, SlabNode.Uninitialized):
                break
            nodes.append(x)

        r = cls._make([*fs, list(nodes)])

        if not (AccountFlag.INITIALIZED in r.account_flags) or (
            AccountFlag.BIDS in r.account_flags == AccountFlag.ASKS in r.account_flags
        ):
            raise ValueError("invalid account_flags for slab")

        return r

    @classmethod
    def from_base64(cls, b: str):
        return cls.from_bytes(base64.b64decode(b))


class Orderbook:
    bids: list[Order]
    asks: list[Order]

    def __init__(self, bids: Slab, asks: Slab, mkt: Market):
        kw = {
            "base_decimals": mkt.base_decimals,
            "quote_decimals": mkt.quote_decimals,
            "base_lot_size": mkt.base_lot_size,
            "quote_lot_size": mkt.quote_lot_size,
        }
        self.bids = [
            Order._make(
                [
                    *o,
                    util.lots_to_price(o.key >> 64, **kw),
                    util.lots_to_size(
                        o.quantity,
                        decimals=mkt.base_decimals,
                        lot_size=mkt.base_lot_size,
                    ),
                    "bid",
                ]
            )
            for o in bids.__reversed__()
        ]
        self.asks = [
            Order._make(
                [
                    *o,
                    util.lots_to_price(o.key >> 64, **kw),
                    util.lots_to_size(
                        o.quantity,
                        decimals=mkt.base_decimals,
                        lot_size=mkt.base_lot_size,
                    ),
                    "ask",
                ]
            )
            for o in asks
        ]



from typing import Literal, Any
from dataclasses import dataclass
from datetime import datetime
from anchorpy import Program
from solana.publickey import PublicKey

Side = Literal["bid", "ask"]
OrderType = Literal[
    "limit", "ioc", "postonly", "reduceonlyioc", "reduceonlylimit", "fok"
]
PerpType = Literal["future", "calloption", "putoption", "square"]


@dataclass(frozen=True)
class CollateralInfo:
    mint: PublicKey
    oracle_symbol: str
    decimals: int
    weight: int
    liq_fee: int
    is_borrowable: bool
    optimal_util: int
    optimal_rate: int
    max_rate: int
    og_fee: int
    is_swappable: bool
    serum_open_orders: PublicKey
    max_deposit: int
    dust_threshold: int
    vault: PublicKey


@dataclass(frozen=True)
class FundingInfo:
    hourly: float
    daily: float
    apr: float


@dataclass(frozen=True)
class MarketInfo:
    address: PublicKey
    symbol: str
    oracle_symbol: str
    perp_type: PerpType
    base_decimals: int
    base_lot_size: int
    quote_decimals: int
    quote_lot_size: int
    strike: int
    base_imf: int
    liq_fee: int
    index_price: float
    mark_price: float
    funding_sample_start_time: datetime
    funding_info: None | FundingInfo

    @property
    def funding_rate(self):
        import warnings

        warnings.warn(
            "Use of deprecated `MarketInfo.funding_rate`, please use `MarketInfo.funding_info`",
            DeprecationWarning,
        )
        return 0 if self.funding_info is None else self.funding_info.hourly


@dataclass(frozen=True)
class PositionInfo:
    size: float
    entry_value: float
    realized_pnl: float
    funding_index: float
    side: Literal["long", "short"]

    @property
    def value(self):
        import warnings

        warnings.warn(
            "Use of deprecated `PositionInfo.value`, please use `PositionInfo.entry_value`",
            DeprecationWarning,
        )
        return self.entry_value


def order_type_from_str(t: OrderType, /, *, program: Program):
    typ = program.type["OrderType"]
    match t:
        case "limit":
            return typ.Limit()
        case "ioc":
            return typ.ImmediateOrCancel()
        case "postonly":
            return typ.PostOnly()
        case "reduceonlyioc":
            return typ.ReduceOnlyIoc()
        case "reduceonlylimit":
            return typ.ReduceOnlyLimit()
        case "fok":
            return typ.FillOrKill()
        case _:
            raise TypeError(f"unsupported order type {t}")


def perp_type_to_str(t: Any, /, *, program: Program) -> PerpType:
    # HACK: Enum comparison is currently broken, so using `str`.
    t = str(t)
    if t == "PerpType.Future()":
        return "future"
    if t == "PerpType.CallOption()":
        return "calloption"
    if t == "PerpType.PutOption()":
        return "putoption"
    if t == "PerpType.Square()":
        return "square"
    raise LookupError(f"invalid perp type {t}")



from typing import *
import math
from anchorpy import Program, Context
from solana.publickey import PublicKey
from solana.keypair import Keypair
from solana.rpc.commitment import Finalized
from solana.rpc.types import TxOpts
from solana.sysvar import SYSVAR_RENT_PUBKEY
import solana.system_program

CONTROL_ACCOUNT_SIZE = 8 + 4482


def decode_symbol(s) -> str:
    s = s.data
    i = s.index(0)
    return bytes(s[:i]).decode("utf-8")


def decode_wrapped_i80f48(n) -> float:
    return n.data / (2**48)


def div_to_float(a: int, b: int) -> float:
    q, r = divmod(a, b)
    gcd = math.gcd(r, b)
    return float(q) + (r // gcd) / (b // gcd)


def big_to_small_amount(n: int | float, /, *, decimals: int) -> int:
    shift = 10 ** abs(decimals)
    if decimals >= 0:
        integral = int(n) * shift
        fractional = int((n % 1) * shift)
        return integral + fractional
    else:
        return int(n) // shift


def small_to_big_amount(n: int | float, /, *, decimals: int):
    return n / 10**decimals


def price_to_lots(
    n: int | float,
    /,
    *,
    base_decimals: int,
    quote_decimals: int,
    base_lot_size: int,
    quote_lot_size: int,
) -> int:
    return round(
        float(n)
        * base_lot_size
        / quote_lot_size
        * 10 ** (quote_decimals - base_decimals)
    )


def lots_to_price(
    n: int,
    /,
    *,
    base_decimals: int,
    quote_decimals: int,
    base_lot_size: int,
    quote_lot_size: int,
) -> float:
    n *= quote_lot_size * 10 ** (base_decimals - quote_decimals)
    return div_to_float(n, base_lot_size)


def size_to_lots(n: float, /, *, decimals: int, lot_size: int) -> int:
    return round(n * 10**decimals) // lot_size


def lots_to_size(n: int, /, *, decimals: int, lot_size: int) -> float:
    return div_to_float(n * lot_size, 10**decimals)


def margin_pda(
    *,
    owner: PublicKey,
    state: PublicKey,
    program_id: PublicKey,
) -> Tuple[PublicKey, int]:
    return PublicKey.find_program_address(
        [
            owner.__bytes__(),
            state.__bytes__(),
            bytes("marginv1", "utf-8"),
        ],
        program_id,
    )


def open_orders_pda(
    *, control: PublicKey, dex_market: PublicKey, program_id: PublicKey
) -> Tuple[PublicKey, int]:
    return PublicKey.find_program_address(
        [control.__bytes__(), dex_market.__bytes__()], program_id
    )


def state_signer_pda(
    *,
    state: PublicKey,
    program_id: PublicKey,
) -> Tuple[PublicKey, int]:
    return PublicKey.find_program_address(
        [
            state.__bytes__(),
        ],
        program_id,
    )


async def create_margin(
    *, program: Program, state: PublicKey, key: PublicKey, nonce: int
) -> str:
    control = Keypair()
    control_lamports = (
        await program.provider.connection.get_minimum_balance_for_rent_exemption(
            CONTROL_ACCOUNT_SIZE
        )
    )["result"]
    return await program.rpc["create_margin"](
        nonce,
        ctx=Context(
            accounts={
                "state": state,
                "authority": program.provider.wallet.public_key,
                "payer": program.provider.wallet.public_key,
                "margin": key,
                "control": control.public_key,
                "rent": SYSVAR_RENT_PUBKEY,
                "system_program": solana.system_program.SYS_PROGRAM_ID,
            },
            pre_instructions=[
                solana.system_program.create_account(
                    solana.system_program.CreateAccountParams(
                        from_pubkey=program.provider.wallet.public_key,
                        new_account_pubkey=control.public_key,
                        lamports=control_lamports,
                        space=CONTROL_ACCOUNT_SIZE,
                        program_id=program.program_id,
                    )
                )
            ],
            signers=[control],
            options=TxOpts(
                max_retries=5,
                preflight_commitment=Finalized,
                skip_confirmation=False,
                skip_preflight=False,
            ),
        ),
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zo']

package_data = \
{'': ['*']}

install_requires = \
['anchorpy>=0.9.3,<0.10.0', 'solana>=0.25.0,<0.26.0']

setup_kwargs = {
    'name': 'zo-sdk',
    'version': '0.1.4',
    'description': '01.xyz Python SDK',
    'long_description': '# 01.xyz Python SDK\n\n<p align="center">\n<b><a href="https://01protocol.github.io/zo-sdk-py/">Documentation</a></b>\n|\n<b><a href="https://pypi.org/project/zo-sdk/">PyPi</a></b>\n</p>\n\nPython SDK to interface with the 01 Solana program.\n\n## Installation\n\n```\n$ pip install zo-sdk\n```\n\n## General Usage\n\n```python\nfrom zo import Zo\n\n# Create the client. By default, this loads the local payer\n# and initializes a margin account for the payer if there\n# isn\'t already one.\nzo = await Zo.new(cluster=\'devnet\')\n\n# View market and collateral info.\nprint(zo.collaterals["BTC"])\nprint(zo.markets["BTC-PERP"])\n\n# Deposit and withdraw collateral.\nawait zo.deposit(1, mint=zo.collaterals["SOL"].mint)\nawait zo.withdraw(1, mint=zo.collaterals["SOL"].mint)\n\n# Place and cancel orders.\nawait zo.place_order(1., 100., \'bid\',\n    symbol="SOL-PERP", order_type="limit", client_id=1)\nawait zo.cancel_order_by_client_id(1, symbol="SOL-PERP")\n\n# Refresh loaded accounts to see updates,\n# such as change in collateral after deposits.\nawait zo.refresh()\n\n# View own balance, positions and orders.\nprint(zo.balance["BTC"])\nprint(zo.position["BTC-PERP"])\nprint(zo.orders["BTC-PERP"])\n\n# Dispatch multiple instructions in a single transaction,\n# using the `_ix` variant.\nawait zo.send(\n    zo.cancel_order_by_client_id_ix(1, symbol="SOL-PERP"),\n    zo.place_order_ix(1., 100., \'bid\',\n        symbol="SOL-PERP", order_type="limit", client_id=1),\n)\n```\n',
    'author': 'Sheheryar Parvaz',
    'author_email': 'me@cherryman.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/01protocol/zo-sdk-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from solana.publickey import PublicKey

from .types import PerpType


@dataclass(kw_only=True)
class Config:
    CLUSTER_URL: str
    ZO_PROGRAM_ID: PublicKey
    ZO_DEX_ID: PublicKey
    ZO_STATE_ID: PublicKey
    SERUM_DEX_ID: PublicKey


configs = {
    "devnet": Config(
        CLUSTER_URL="https://api.devnet.solana.com",
        ZO_PROGRAM_ID=PublicKey("Zo1ThtSHMh9tZGECwBDL81WJRL6s3QTHf733Tyko7KQ"),
        ZO_DEX_ID=PublicKey("ZDxUi178LkcuwdxcEqsSo2E7KATH99LAAXN5LcSVMBC"),
        ZO_STATE_ID=PublicKey("KwcWW7WvgSXLJcyjKZJBHLbfriErggzYHpjS9qjVD5F"),
        SERUM_DEX_ID=PublicKey("DESVgJVGajEgKGXhb6XmqDHGz3VjdgP7rEVESBgxmroY"),
    ),
    "mainnet": Config(
        CLUSTER_URL="https://api.mainnet-beta.solana.com",
        ZO_PROGRAM_ID=PublicKey("Zo1ggzTUKMY5bYnDvT5mtVeZxzf2FaLTbKkmvGUhUQk"),
        ZO_DEX_ID=PublicKey("ZDx8a8jBqGmJyxi1whFxxCo5vG6Q9t4hTzW2GSixMKK"),
        ZO_STATE_ID=PublicKey("71yykwxq1zQqy99PgRsgZJXi2HHK2UDx9G4va7pH6qRv"),
        SERUM_DEX_ID=PublicKey("9xQeWvG816bUx9EPjHmaT23yvVM2ZWbrrpZb9PusVFin"),
    ),
}


def taker_fee(t: PerpType, /) -> float:
    if t == "future":
        return 10 / 10000
    if t == "calloption" or t == "putoption":
        return 10 / 10000
    if t == "square":
        return 15 / 10000
    raise LookupError(f"invalid perp type {t}")
